// BEGIN FLOCK GPL
//
// Copyright Flock Inc. 2005-2008
// http://flock.com
//
// This file may be used under the terms of the
// GNU General Public License Version 2 or later (the "GPL"),
// http://www.gnu.org/licenses/gpl.html
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// END FLOCK GPL

const CC = Components.classes;
const CI = Components.interfaces;
const CR = Components.results;
const CU = Components.utils;

CU.import("resource:///modules/FlockXPCOMUtils.jsm");
FlockXPCOMUtils.debug = false;

CU.import("resource:///modules/FlockSvcUtils.jsm");
CU.import("resource:///modules/FlockStringBundleHelpers.jsm");
CU.import("resource:///modules/FlockXMLUtils.jsm");
CU.import("resource://gre/modules/JSON.jsm");
CU.import("resource:///modules/webmailBackend.jsm");

const MODULE_NAME = "AOLMail";
const CLASS_NAME = "Flock AOL Mail Service";
const CLASS_SHORT_NAME = "aolmail";
const CLASS_TITLE = "AOL Mail";
const CLASS_ID = Components.ID("{81690132-CA72-4041-AD8E-27FC6AA1337A}");
const CONTRACT_ID = "@flock.com/aolmail-service;1";

const API_CLASS_NAME = "AOLMail API";
const API_CLASS_ID = Components.ID("{22A698AF-46DD-44AB-93EB-4769189E5CB6}");
const API_CONTRACT_ID = "@flock.com/webservice/api/aolmail;1";

const SERVICE_ENABLED_PREF = "flock.service.aolmail.enabled";
const DISPLAY_UNREAD_COUNT_PREF = "flock.webmail.displayUnreadMessageCount";

const AOLMAIL_URL = "http://webmail.aol.com";
const AOLMAIL_FAVICON = "chrome://flock/content/services/aolmail/favicon.png";
const AOLMAIL_STRING_BUNDLE = "chrome://flock/locale/services/aolmail.properties";

const FLOCK_ERROR_CONTRACTID = "@flock.com/error;1";
const PREFS = CC["@mozilla.org/preferences-service;1"]
              .getService(CI.nsIPrefBranch);

const XMLHTTPREQUEST_CONTRACTID = "@mozilla.org/xmlextras/xmlhttprequest;1";
const XMLHTTPREQUEST_READYSTATE_COMPLETED = 4;

const STR_EMAIL = "email";
const STR_ISBASICVERSION = "isBasicVersion";
const STR_ISUTF16 = "isUTF16Encoded";
const STR_LANGUAGE = "language";
const STR_LOCALE = "locale";
const STR_UNREAD = "unreadMessages";
const STR_WAITING_TO_SEND = "waitingToSend";
const STR_API_TOKEN = "apiToken";
const STR_API_EXPIRES_ON = "apiExpiresOn";

const DOMAIN = "aol.com";
const MAX_UNREAD_COUNT = 75;  // max unread messages to retrieve
const REFRESH_INTERVAL = 600;  // 10 min in seconds
const MAX_DISPLAY_COUNT = 20;
const DEFAULT_DISPLAY_COUNT = 4;
const MORE_AUTHENTICATION_REQUIRED_CODE = 330;

var gApi = null;


/**************************************************************************
 * Component: Flock AOLMailService
 **************************************************************************/

function AOLMailService() {
  this._init();

  var ws = FlockSvcUtils.flockIWebService;
  ws.addDefaultMethod(this, "getAccount");
  ws.addDefaultMethod(this, "getAuthenticatedAccount");
  ws.addDefaultMethod(this, "getAccounts");
  ws.addDefaultMethod(this, "getStringBundle");
  ws.addDefaultMethod(this, "isHidden");
  ws.addDefaultMethod(this, "logout");
  this._loginURL = AOLMAIL_URL;
  ws.addDefaultMethod(this, "loginURL");

  var lws = FlockSvcUtils.flockIManageableWebService;
  lws.addDefaultMethod(this, "getCredentialsFromForm");
  lws.addDefaultMethod(this, "getSessionValue");
  lws.addDefaultMethod(this, "ownsDocument");
  lws.addDefaultMethod(this, "ownsLoginForm");
}


/**************************************************************************
 * Flock AOLMailService: XPCOM Component Creation
 **************************************************************************/

AOLMailService.prototype = new FlockXPCOMUtils.genericComponent(
  CLASS_NAME,
  CLASS_ID,
  CONTRACT_ID,
  AOLMailService,
  CI.nsIClassInfo.SINGLETON,
  [
    CI.nsIObserver,
    CI.nsITimerCallback,
    CI.flockIPollingService,
    CI.flockIWebService,
    CI.flockIManageableWebService,
    CI.flockIMailWebService
  ]
);

// FlockXPCOMUtils.genericModule() categories
AOLMailService.prototype._xpcom_categories = [
  { category: "wsm-startup" },
  { category: "flockWebService", entry: CLASS_SHORT_NAME }
];


/**************************************************************************
 * Flock AOLMailService: Private Data and Methods
 **************************************************************************/

// Service initialization
AOLMailService.prototype._init =
function amSvc__init() {
  FlockSvcUtils.getLogger(this).info(".init()");

  // Determine whether this service has been disabled, and unregister if so
  if (PREFS.getPrefType(SERVICE_ENABLED_PREF) &&
      !PREFS.getBoolPref(SERVICE_ENABLED_PREF))
  {
    this._logger.info("Pref " + SERVICE_ENABLED_PREF
                              + " set to FALSE... not initializing.");
    this.deleteCategories();
    return;
  }

  gApi = new AOLMailAPI();

  profiler = CC["@flock.com/profiler;1"].getService(CI.flockIProfiler);
  var evtID = profiler.profileEventStart("aolmail-init");

  var obs = CC["@mozilla.org/observer-service;1"]
            .getService(CI.nsIObserverService);
  obs.addObserver(this, "xpcom-shutdown", false);

  this._accountClassCtor = AOLMailAccount;
  FlockSvcUtils.getWD(this);
  FlockSvcUtils.getCoop(this);
  if (this._coop.Service.exists(this.urn)) {
    this._c_svc = this._coop.get(this.urn);
  } else {
    this._c_svc = new this._coop.Service(this.urn, {
      name: CLASS_SHORT_NAME,
      desc: CLASS_TITLE
    });
  }
  this._c_svc.serviceId = CONTRACT_ID;
  this._c_svc.logoutOption = true;
  this._c_svc.domains = this._WebDetective
                            .getString(CLASS_SHORT_NAME, "domains", DOMAIN);

  // Update auth states
  try {
    if (this._WebDetective
            .detectCookies(CLASS_SHORT_NAME, "loggedoutCookies", null))
    {
      this._acUtils.markAllAccountsAsLoggedOut(CONTRACT_ID);
    }
    gApi.deauthenticate();
  } catch (ex) {
    this._logger.error("ERROR updating auth states for "
                       + MODULE_NAME
                       + ": "
                       + ex);
  }

  // Initialize member that specifies path for localized string bundle
  this._stringBundlePath = AOLMAIL_STRING_BUNDLE;

  profiler.profileEventEnd(evtID, "");
}

AOLMailService.prototype._refreshUnreadMessages =
function amSvc__refreshUnreadMessages(aAccount, aListener) {
  this._logger.debug("_refreshUnreadMessages()");
  aListener.onStart(this, "_refreshUnreadMessages");
  var svc = this;
  var newMailListener = {
    onStart: function rum_newMailListener_onStart(aSubj, aTopic) {
      svc._logger.debug("newMailListener.onStart()");
    },
    onSuccess: function rum_newMailListener_onSuccess(aResult, aTopic) {
      svc._logger.debug("newMailListener.onSuccess()");
      var messageList = aResult.wrappedJSObject.mailList.messageList;
      var getLink = function rum_getLink(aLink) {
        var link = aLink;
        // Return adjusted link if account is using basic version
        if (aAccount.getCustomParam(STR_ISBASICVERSION)) {
          // Get message id from link
          var uid_re = svc._WebDetective.getString(svc.shortName, "uid_re", "");
          var getId = link.match(uid_re);
          if (getId) {
            // Build basic link
            link = svc._WebDetective
                      .getString(svc.shortName, "basicMessageURL", "")
                      .replace("%id%", getId[1]);
          }
        }
        return link;
      };
      // Get the messages
      var newCount = 0;
      var msgCount = 0;
      var messages = [];
      var newestMessageDate = 0;
      for (var i = 0; i < messageList.length; i++) {
        var folder = messageList[i];
        if (folder.folderName.match(/inbox/gi)) {
          newCount = folder.totalCount;
          msgCount = folder.messageCount;
          for (var j = 0; j < msgCount; j++) {
            messages.push({
              from: folder.messages[j].sender,
              subject: folder.messages[j].subject,
              url: getLink(folder.messages[j].messageLink)
            });
            // Retrieve the date of the newest message
            if (!newestMessageDate) {
              var datetime_re = svc._WebDetective
                                   .getString(svc.shortName, "datetime_re", "");
              var sentOn = folder.messages[j].sentOn.match(datetime_re);
              if (sentOn) {
                // Parse the DateTime field.
                // Format example: 2008-02-26T19:32:19-05:00
                var year = sentOn[1];
                var month = parseInt(sentOn[2], 10) - 1;
                var date = parseInt(sentOn[3], 10);
                var hrs = parseInt(sentOn[4], 10);
                var min = parseInt(sentOn[5], 10);
                var sec = parseInt(sentOn[6], 10);
                var tzHrs = parseInt(sentOn[8], 10);
                var tzMin = parseInt(sentOn[9], 10);
                if (sentOn[7] == "-") {
                  hrs += tzHrs;
                  min += tzMin;
                } else {
                  hrs -= tzHrs;
                  min -= tzMin;
                }
                newestMessageDate = Date.UTC(year, month, date, hrs, min, sec);
              }
            }
          }
          break;
        }
      }

      var unreadMessages = {
        lastCheckedDate: Date.now(),
        newestMessageDate: newestMessageDate,
        unreadMessageCount: newCount,
        messages: messages
      };

      aListener.onSuccess(JSON.toString(unreadMessages),
                          "_refreshUnreadMessages");
    },
    onError: function rum_newMailListener_onError(aSubj, aTopic, aError) {
      svc._logger.error("newMailListener.onError('" + aError.errorCode + "')");
      aListener.onError(svc, "_refreshUnreadMessages", aError);
    }
  };

  var numDisplayed = 0;
  if (PREFS.getPrefType(DISPLAY_UNREAD_COUNT_PREF)) {
    numDisplayed = PREFS.getIntPref(DISPLAY_UNREAD_COUNT_PREF);
  }
  if (!numDisplayed || numDisplayed > MAX_DISPLAY_COUNT) {
    numDisplayed = DEFAULT_DISPLAY_COUNT;
  }

  gApi.getNewMailList(aAccount.username,
                      "Inbox",
                      numDisplayed,
                      true,
                      aAccount.getCustomParam(STR_LANGUAGE),
                      aAccount.getCustomParam(STR_LOCALE),
                      aAccount,
                      newMailListener);
}


/**************************************************************************
 * Flock AOLMailService: flockIWebService Implementation
 **************************************************************************/

// readonly attribute AString title;
AOLMailService.prototype.title = CLASS_TITLE;

// readonly attribute AString shortName;
AOLMailService.prototype.shortName = CLASS_SHORT_NAME;

// readonly attribute AString icon;
AOLMailService.prototype.icon = AOLMAIL_FAVICON;

// readonly attribute AString url;
AOLMailService.prototype.url = AOLMAIL_URL;

// readonly attribute AString contractId;
// ALMOST duplicated by nsIClassInfo::contractID
// with different case: contractId != contractID
// FIXME: File a bug for this as IDL is case-insensitive.
AOLMailService.prototype.contractId = CONTRACT_ID;

// readonly attribute AString urn;
AOLMailService.prototype.urn = "urn:" + CLASS_SHORT_NAME + ":service";

// DEFAULT: nsIStringBundle getStringBundle();
// DEFAULT: boolean isHidden();


/**************************************************************************
 * Flock AOLMailService: flockIManageableWebService Implementation
 **************************************************************************/

// readonly attribute AString domains;
AOLMailService.prototype.__defineGetter__("domains",
function amSvc_getDomains() {
  this._logger.debug("Getting property: domains");
  return this._c_svc.domains;
});

// readonly attribute boolean needPassword;
AOLMailService.prototype.needPassword = true;

// addAccount(in AString aAccountID,
//            in boolean aIsTransient,
//            in flockIListener aListener);
AOLMailService.prototype.addAccount =
function amSvc_addAccount(aAccountID, aIsTransient, aListener) {
  this._logger.debug("addAccount('" + aAccountID + "', "
                                    + aIsTransient + ", aListener)");

  if (!aAccountID) {
    if (aListener) {
      var error = CC[FLOCK_ERROR_CONTRACTID].createInstance(CI.flockIError);
      // XXX See bug 10749 - flockIError cleanup
      error.errorCode = 9999;
      error.errorString = "No Account ID provided";
      aListener.onError(this, "addAccount", error);
    }
    return null;
  }

  var pw = this._acUtils.getPassword(this.urn + ":" + aAccountID);
  var user = pw ? pw.user : aAccountID;

  // Create a new account and add it to coop if it doesn't already exist
  var accountUrn = "urn:flock:" + this.shortName + ":account:" + aAccountID;
  if (!this._coop.Account.exists(accountUrn)) {
    var coopAccount = new this._coop.Account(accountUrn, {
      accountId: aAccountID,
      favicon: this.icon,
      isPollable: true,
      isTransient: aIsTransient,
      name: user,
      refreshInterval: REFRESH_INTERVAL,
      service: this._c_svc,
      serviceId: this.contractId
    });
    this._coop.accounts_root.children.addOnce(coopAccount);
  }

  // Add custom parameters to be used
  var params = CC["@mozilla.org/hash-property-bag;1"]
               .createInstance(CI.nsIWritablePropertyBag2);
  params.setPropertyAsAString(STR_EMAIL, "");
  params.setPropertyAsBool("isPrimary", false);
  params.setPropertyAsBool("hasNewMessages", false);
  params.setPropertyAsBool(STR_ISUTF16, false);
  params.setPropertyAsInt32("lastPrimaryDate", 0);
  params.setPropertyAsBool(STR_ISBASICVERSION, false);
  params.setPropertyAsAString(STR_LANGUAGE, "en");
  params.setPropertyAsAString(STR_LOCALE, "us");
  params.setPropertyAsAString(STR_UNREAD,
                              webmailBackend.createEmptyMessageAsJSON());
  params.setPropertyAsAString(STR_WAITING_TO_SEND, "");
  params.setPropertyAsAString(STR_API_TOKEN, "");
  params.setPropertyAsInt32(STR_API_EXPIRES_ON, 0);
  this._acUtils.addCustomParamsToAccount(params, accountUrn);

  // Instantiate account component
  var account = this.getAccount(accountUrn);
  if (aListener) {
    aListener.onSuccess(account, "addAccount");
  }
  return account;
}

// boolean docRepresentsSuccessfulLogin(in nsIDOMHTMLDocument aDocument);
AOLMailService.prototype.docRepresentsSuccessfulLogin =
function amSvc_docRepresentsSuccessfulLogin(aDocument) {
  this._logger.debug("docRepresentsSuccessfulLogin()");
  if (this._WebDetective.detect(this.shortName, "loggedin", aDocument, null)) {
    // We're logged in but can we get an account ID yet?
    if (this.getAccountIDFromDocument(aDocument)) {
      // Got the account ID
      return true;
    }
    // No account ID, but the page might be in flux due to scripts and ajax
    // events. Monitor the page a bit longer and see if the account ID shows up.
    // Note: we cannot call asyncDetect() here because we do not have access to
    // aDocument's container element.
    var svc = this;
    var timerCB = {
      tries: 0,
      notify: function am_drsl_notify(aTimer) {
        this.tries++;
        if (this.tries > 50) {
          // Didn't turn up
          aTimer.cancel();
        } else if (svc._WebDetective
                      .detect(svc.shortName, "accountinfo", aDocument, null))
        {
          // Success!
          aTimer.cancel();
          // Hack: Now that we know the page has loaded the info we need to
          // extract from it, we issue another "FlockDocumentReady" event. This
          // will be picked up by the observer in flockWebServiceManager, which
          // will in turn call this method again.
          var obs = CC["@mozilla.org/observer-service;1"]
                    .getService(CI.nsIObserverService);
          obs.notifyObservers(aDocument,
                              "FlockDocumentReady",
                              aDocument.location.href);
         }
      }
    };
    var asyncTimer = CC["@mozilla.org/timer;1"].createInstance(CI.nsITimer);
    asyncTimer.initWithCallback(timerCB, 100, CI.nsITimer.TYPE_REPEATING_SLACK);
  }
  return false;
}

// AString getAccountIDFromDocument(in nsIDOMHTMLDocument aDocument);
AOLMailService.prototype.getAccountIDFromDocument =
function amSvc_getAccountIDFromDocument(aDocument) {
  this._logger.debug("getAccountIDFromDocument()");
  var results = FlockSvcUtils.newResults();
  if (this._WebDetective
          .detect(this.shortName, "accountinfo", aDocument, results))
  {
    var matches = results.getPropertyAsAString("accountid")
                         .match(/\s*(\S+\w)\W*$/m);
    if (matches) {
      // Get the account name. Note: matches[1] will not be null.
      var accountName = matches[1];
      // Is it a full email address?
      var email = accountName.match(/(.+)@(.+)/);
      if (email) {
        // If the email address is from one of the AOL domains, strip off
        // everything after and including the "@".
        var domains = this.domains.split(",");
        for each (var domain in domains) {
          // Note: email[1] and email[2] will not be null.
          if (domain == email[2]) {
            // Found common domain so return account ID without "@domain". 
            return email[1];
          }
        }
      }
      // Not an email address or from an AOL domain, so use the full account
      // name as the ID.
      return accountName;
    }
  }
  return null;
}

// DEFAULT: nsISimpleEnumerator getAccounts();
// DEFAULT: flockIWebServiceAccount getAccount(in AString aUrn);
// DEFAULT: flockIWebServiceAccount getAuthenticatedAccount();
// DEFAULT: nsIPassword getCredentialsFromForm(in nsIDOMHTMLFormElement aForm);
// DEFAULT: AString getSessionValue();
// DEFAULT: void logout();
// DEFAULT: boolean ownsDocument(in nsIDOMHTMLDocument aDocument);
// DEFAULT: boolean ownsLoginForm(in nsIDOMHTMLFormElement aForm);

// void removeAccount(in AString aUrn);
AOLMailService.prototype.removeAccount =
function amSvc_removeAccount(aUrn) {
  this._logger.debug("removeAccount('" + aUrn + "')");
  this._acUtils.removeAccount(aUrn);
  webmailBackend.ensurePrimaryAccountExists();
}

// void updateAccountStatusFromDocument(in nsIDOMHTMLDocument aDocument,
//                                      in AString aAcctURN);
AOLMailService.prototype.updateAccountStatusFromDocument =
function amSvc_updateAccountStatusFromDocument(aDocument,
                                               aAcctURN)
{
  this._logger.debug("updateAccountStatusFromDocument('" + aAcctURN + "')");
  if (aAcctURN) {
    // We are currently logged in to aAcctURN
    var account = this.getAccount(aAcctURN);
    if (!this._acUtils.ensureOnlyAuthenticatedAccount(aAcctURN)) {
      // Just logged in
      webmailBackend.onLogin(account);
    }
    // Pull information off of the webmail page
    var url = aDocument.location;
    // Get version
    var basicFlag = this._WebDetective.getString(this.shortName, "basic", "");
    var isBasic = (url.pathname.indexOf(basicFlag) > -1);
    account.setCustomParam(STR_ISBASICVERSION, isBasic);
    account.setCustomParam(STR_ISUTF16, isBasic);
    // Get language and locale
    var matched = url.pathname.match(/\/([a-z][a-z])-([a-z][a-z])\//i);
    if (matched) {
      account.setCustomParam(STR_LANGUAGE, matched[1]);
      account.setCustomParam(STR_LOCALE, matched[2]);
    }

    // Store the email address, if necessary
    account._storeEmail(aDocument);

    // Do we have a pending compose message?
    webmailBackend.checkForPendingComposeMessage(account);
  } else if (this._WebDetective
                 .detect(this.shortName, "loggedout", aDocument, null))
  {
    // Detected a logged out state
    this._acUtils.markAllAccountsAsLoggedOut(CONTRACT_ID);
    gApi.deauthenticate();

    // Ensure the webmail icon does not indicate new unread messages while
    // logged out of the primary account
    webmailBackend.updateIcon();
  }
}


/**************************************************************************
 * Flock AOLMailService: flockIPollingService Implementation
 **************************************************************************/

// void refresh(in AString aUrn, in flockIPollingListener aListener);
AOLMailService.prototype.refresh =
function amSvc_refresh(aURN, aPollingListener) {
  this._logger.debug("refresh(" + aURN + ")");

  var account = this.getAccount(aURN);
  if (!account.isAuthenticated()) {
    return;
  }
  var svc = this;
  var listener = {
    onStart: function refresh_onStart(aSubj, aTopic) {
      svc._logger.debug("Start");
    },
    onSuccess: function refresh_onSuccess(aMsgs, aTopic) {
      svc._logger.debug("Success");
      webmailBackend.handleRefreshedMessages(svc, aURN, aMsgs);
      aPollingListener.onResult();
    },
    onError: function refresh_onError(aSubj, aTopic, aError) {
      svc._logger.debug("unable to refresh account " + aURN);
      svc.getAccount(aURN)
         .setCustomParam(STR_UNREAD,
                         webmailBackend.createEmptyMessageAsJSON());
      aPollingListener.onError(null);
    }
  }
  this._refreshUnreadMessages(account, listener);
}


/**************************************************************************
 * Flock AOLMailService: nsIObserver Implementation
 **************************************************************************/

// void observe(nsISupports subject, char* topic, PRUnichar* data);
AOLMailService.prototype.observe =
function amSvc_observe(aSubject, aTopic, aState) {
  switch (aTopic) {
    case "xpcom-shutdown":
      var obs = CC["@mozilla.org/observer-service;1"]
                .getService(CI.nsIObserverService);
      obs.removeObserver(this, "xpcom-shutdown");
      break;
  }
}


/**************************************************************************
 * Component: Flock AOLMailAccount
 **************************************************************************/

// Constructor
function AOLMailAccount() {
  FlockSvcUtils.getLogger(this).init(CLASS_SHORT_NAME + ":Account");
  this._logger.info(".init()");
  FlockSvcUtils.getACUtils(this);
  FlockSvcUtils.getCoop(this);
  // TODO XXX: Should be able to use FlockSvcUtils.getWD() here, but it can
  //           only be called by the service.  Need to fix.
  this._WebDetective = CC["@flock.com/web-detective;1"]
                       .getService(CI.flockIWebDetective);

  var wsa = FlockSvcUtils.flockIWebServiceAccount;
  wsa.addDefaultMethod(this, "deactivate");
  wsa.addDefaultMethod(this, "getService");
  wsa.addDefaultMethod(this, "isAuthenticated");
  wsa.addDefaultMethod(this, "getParam");
  wsa.addDefaultMethod(this, "setParam");
  wsa.addDefaultMethod(this, "getCustomParam");
  wsa.addDefaultMethod(this, "getAllCustomParams");
  wsa.addDefaultMethod(this, "setCustomParam");

  // TODO XXX: Setting a private _shortName when adding the
  //           flockIWebmailAccount default methods should only be a
  //           temporary measure until the account has reference to its service
  this._shortName = CLASS_SHORT_NAME;
  var wma = FlockSvcUtils.flockIMailWebServiceAccount;
  wma.addDefaultMethod(this, "getUnreadMessages");
  wma.addDefaultMethod(this, "getUpgradeAccount");
  wma.addDefaultMethod(this, "isPrimary");
  wma.addDefaultMethod(this, "refreshUnreadMessageList");
  wma.addDefaultMethod(this, "setAsPrimary");
}


/**************************************************************************
 * Flock AOLMailAccount: XPCOM Component Creation
 **************************************************************************/

// Use genericComponent() for the goodness it provides (QI, nsIClassInfo, etc),
// but do NOT add AOLMailAccount to the list of constructors passed
// to FlockXPCOMUtils.genericModule().
AOLMailAccount.prototype = new FlockXPCOMUtils.genericComponent(
  CLASS_NAME + " Account",
  "",
  "",
  AOLMailAccount,
  0,
  [
    CI.flockIWebServiceAccount,
    CI.flockIMailWebServiceAccount
  ]
);


/**************************************************************************
 * Flock AOLMailAccount: flockIWebServiceAccount Implementation
 **************************************************************************/

// readonly attribute AString urn;
AOLMailAccount.prototype.urn = "";

// void activate(in flockIListener aListener);
AOLMailAccount.prototype.activate =
function amAcct_activate(aListener) {
  this._logger.debug("activate()");
  this._coop.get(this.urn).isAuthenticated = true;
  webmailBackend.onLogin(this);

  if (aListener) {
    aListener.onSuccess(this, "activate");
  }
}

// DEFAULT: flockILoginWebService getService();

// void login(in flockIListener aListener);
AOLMailAccount.prototype.login =
function amAcct_login(aListener) {
  this._logger.debug("login()");
  if (aListener) {
    aListener.onStart(this, "login");
  }
  var inst = this;
  var authListener = {
    onStart: function login_authListener_onStart(aSubj, aTopic) {
      inst._logger.debug("login_authListeneronStart()");
    },
    onSuccess: function login_authListener_onSuccess(aSubj, aTopic) {
      inst._logger.debug("login_authListeneronSuccess()");
      var c_acct = inst._coop.get(inst.urn);
      if (!c_acct.isAuthenticated) {
        c_acct.isAuthenticated = true;
        c_acct.isPollable = true;
      }
      if (aListener) {
        aListener.onSuccess(inst, "login");
      }
    },
    onError: function login_authListener_onError(aSubj, aTopic, aError) {
      inst._logger.error("login_authListener_onError()");
      if (aListener) {
        aListener.onError(inst, "login", aError);
      }
    }
  };

  // Authenticate the API
  gApi._authenticate(this, null, authListener);
}

// void logout(in flockIListener aListener);
AOLMailAccount.prototype.logout =
function amAcct_logout(aListener) {
  this._logger.debug("logout()");
  var c_acct = this._coop.get(this.urn);
  if (c_acct.isAuthenticated) {
    c_acct.isAuthenticated = false;
    gApi.deauthenticate();
    webmailBackend.updateIcon();
    CC[c_acct.serviceId].getService(CI.flockIWebService).logout();
  }
  if (aListener) {
    aListener.onSuccess(this, "logout");
  }
}

// void keep();
AOLMailAccount.prototype.keep =
function amAcct_keep() {
  this._logger.debug("keep()");
  this._coop.get(this.urn).isTransient = false;
  var urn = this.urn.replace("account:", "service:").replace("flock:", "");
  this._acUtils.makeTempPasswordPermanent(urn);

  // Automatically set the webmail account as the default mail client on the
  // first time the user configures a webmail account
  if (PREFS.getPrefType("flock.webmail.promptAsDefaultMailer") &&
      PREFS.getBoolPref("flock.webmail.promptAsDefaultMailer"))
  {
    PREFS.setBoolPref("flock.webmail.promptAsDefaultMailer", false);
    PREFS.setBoolPref("flock.webmail.useAsDefaultMailer", true);
  }
}

// DEFAULT: void isAuthenticated();
// DEFAULT: nsIVariant getParam(in AString aParamName);
// DEFAULT: void setParam(in AString aParamName, in nsIVariant aValue);
// DEFAULT: nsIVariant getCustomParam(in AString aParamName);
// DEFAULT: nsIPropertyBag getAllCustomParams();
// DEFAULT: void setCustomParam(in AString aParamName, in nsIVariant aValue);


/**************************************************************************
 * Flock AOLMailAccount: flockIWebmailAccount Implementation
 **************************************************************************/

// AString getComposeURL(in AString aMailParams);
AOLMailAccount.prototype.getComposeURL =
function amAcct_getComposeURL(aMailParams) {
  this._logger.debug("getComposeURL('" + aMailParams + "')");
  var version = this._getVersionString();
  var params = webmailBackend.reformatMailParams(this, aMailParams, ",");
  if (params && version == "standard") {
    // The "standard" AOL Mail dies if there are newlines and carriage returns
    // in the params, so we need to replace them. We use 0x08, framed by
    // spaces, as it is very unlikely to encounter this sequences in a mailto
    // string (and if we did, we'd want to replace it regardless). The spaces
    // ensure that the 0x08 is not "bound" to its neighbors when the composer
    // parses the content. These new values will get replaced when we call
    // onComposeURLLoaded() later on. 
    params = params.replace(/%0a/gi, "%20%08%20")
                   .replace(/%0d/gi, "%20%08%20");
  }

  // Get URL and add params
  var url = this._WebDetective
                .getString(CLASS_SHORT_NAME, version + "ComposeURL", "")
                .replace("%params%", params);
  this._logger.debug("compose URL is: " + url);
  return url;
}

// void onComposeURLLoaded(in AString aComposeURL,
//                         in nsIDOMXULElement aTabBrowser);
AOLMailAccount.prototype.onComposeURLLoaded =
function amAcct_onComposeURLLoaded(aComposeURL, aTabBrowser) {
  this._logger.debug("onComposeURLLoaded()");
  // There are two issues to deal with after the compose page has loaded:
  // 1. In "standard", we may have replaced linebreaks and carriage returns
  //    with "%20%08%20", so we have to look for those in the message body and
  //    replace with "<br >".
  // 2. In "basic", the body parameter needs to be added after the page loads.
  //    The standard "body=" param doesn't work, however, "richbody" seems to
  //    work for writing to the body of the rich text editor. I cannot find an
  //    equivalent for the plain text editor.
  if (this.getCustomParam(STR_ISBASICVERSION)) {
    this._onComposeURLLoadedBasic(aComposeURL, aTabBrowser);
  } else {
    this._onComposeURLLoadedStandard(aComposeURL, aTabBrowser);
  }
}

// AString getEmailAddress();
AOLMailAccount.prototype.getEmailAddress =
function amAcct_getEmailAddress() {
  this._logger.debug("getEmailAddress()");
  var email = this.getCustomParam(STR_EMAIL);
  if (!email) {
    email = this._coop.get(this.urn).accountId;
  }
  this._logger.debug("email address is: " + email);
  return email;
}

// AString getInboxURL();
AOLMailAccount.prototype.getInboxURL =
function amAcct_getInboxURL() {
  var url = this._WebDetective
                .getString(CLASS_SHORT_NAME,
                           this._getVersionString() + "InboxURL",
                           "");
  this._logger.debug("getInboxURL(): " + url);
  return url;
}

// DEFAULT: AString getUnreadMessages(in boolean aCheckNow);
// DEFAULT: AString getUpgradeAccount();
// DEFAULT: boolean isPrimary();
// DEFAULT: void refreshUnreadMessageList();
// DEFAULT: void setAsPrimary();


/**************************************************************************
 * Flock AOLMailAccount: Private Data and Methods
 **************************************************************************/

AOLMailAccount.prototype._onComposeURLLoadedBasic =
function amAcct__onComposeURLLoadedBasic(aComposeURL, aTabBrowser) {
  this._logger.debug("_onComposeURLLoadedBasic()");
  // We need to insert the message body into the message after the page has
  // loaded. Parse the mailto parameters out of the compose URL to get the body.
  var params = webmailBackend.parseMailParams(this, aComposeURL);
  if (!params.body) {
    // No body, so we're done
    return;
  }
  var inst = this;  // Wait for the text editor to load
  var results = FlockSvcUtils.newResults();
  var composerListener = {
    onStart: function bv_composeListener_onStart(aSubj, aTopic) {
      inst._logger.debug("bv_composerListener: onStart()");
    },
    onSuccess: function bv_composeListener_onSuccess(aSubj, aTopic) {
      inst._logger.info("bv_composerListener: onSuccess()");

      // What loaded: Plain text or Rich text editor?
      var editorNode = results.getPropertyAsInterface("node", CI.nsIDOMNode);
      if (editorNode instanceof CI.nsIDOMHTMLTextAreaElement) {
        // Plain Text editor. We prepend the body data to the TEXTAREA data, in
        // case the user has a signature
        editorNode.value = params.body + editorNode.value;
        return;
      }

      // Rich Text editor
      if (editorNode instanceof CI.nsIDOMHTMLIFrameElement) {
        var bodyListener = {
          onStart: function bv_bodyListener_onStart(aSubj, aTopic) {
            inst._logger.debug("bv_bodyListener: onStart()");
          },
          onSuccess: function bv_bodyListener_onSuccess(aSubj, aTopic) {
            inst._logger.info("bv_bodyListener: onSuccess()");
            var body = editorNode.contentDocument.body;
            // Replace the newlines with breaks. We prepend the body data to the
            // message body, in case the user has a signature
            body.innerHTML = params.body.replace(/\x0A/g, "<br />")
                           + body.innerHTML;
          },
          onError: function bv_bodyListener_onError(aSubj, aTopic, aError) {
            inst._logger.error("bv_bodyListener: onError()");
          }
        };
        // Wait for the body to get filled
        inst._WebDetective
            .asyncDetect(CLASS_SHORT_NAME, "basicRichComposerBody",
                         editorNode, null, bodyListener, 250, 80);
      }
    },
    onError: function bv_composerListener_onError(aSubj, aTopic, aError) {
      inst._logger.error("bv_composerListener: onError()");
    }
  };
  
  this._WebDetective
      .asyncDetect(CLASS_SHORT_NAME, "basicComposer",
                   aTabBrowser, results, composerListener, 250, 80);
}

AOLMailAccount.prototype._onComposeURLLoadedStandard =
function amAcct__onComposeURLLoadedStandard(aComposeURL, aTabBrowser) {
  this._logger.debug("_onComposeURLLoadedStandard()");
  // The Standard version cannot handle newlines in the compose URL so we need
  // to pound them into the message body after it has loaded, replacing the
  // "%20%08%20" sequences we inserted in getComposeURL() with "<br />".
  var inst = this;
  var results = FlockSvcUtils.newResults();
  var composerListener = {
    onStart: function nv_composerListener_onStart(aSubj, aTopic) {
      inst._logger.debug("nv_composerListener: onStart()");
    },
    onSuccess: function nv_composeListener_onSuccess(aSubj, aTopic) {
      inst._logger.info("nv_composerListener: onSuccess()");

      // What loaded: Plain text or Rich text editor?
      var editorNode = results.getPropertyAsInterface("node", CI.nsIDOMNode);
      if (editorNode instanceof CI.nsIDOMHTMLTextAreaElement) {
        // Plain Text editor. Restore newlines.
        editorNode.value = editorNode.value.replace(/ ?\x08 ?/g, "\n");
        return;
      }

      // Rich Text editor
      if (editorNode instanceof CI.nsIDOMHTMLIFrameElement) {
        var bodyListener = {
          onStart: function bv_bodyListener_onStart(aSubj, aTopic) {
            inst._logger.debug("bv_bodyListener: onStart()");
          },
          onSuccess: function bv_bodyListener_onSuccess(aSubj, aTopic) {
            inst._logger.info("bv_bodyListener: onSuccess()");
            var body = editorNode.contentDocument.body;
            // Replace the newlines with breaks
            body.innerHTML = body.innerHTML.replace(/ ?\x08 ?/g, "<br />");
          },
          onError: function bv_bodyListener_onError(aSubj, aTopic, aError) {
            inst._logger.error("bv_bodyListener: onError()");
          }
        };
        // Wait for the body to get filled
        inst._WebDetective
            .asyncDetect(CLASS_SHORT_NAME, "standardRichComposerBody",
                         editorNode, null, bodyListener, 250, 80);
      }
    },
    onError: function nv_composerListener_onError(aSubj, aTopic, aError) {
      inst._logger.error("nv_composerListener: onError()");
    }
  };
  // Wait for AOL Mail to load within the "ws_hat" shell
  this._WebDetective
      .asyncDetect(CLASS_SHORT_NAME, "standardComposer",
                   aTabBrowser, results, composerListener, 250, 80);
}

AOLMailAccount.prototype._getVersionString =
function amAcct__getVersionString() {
  return this.getCustomParam(STR_ISBASICVERSION) ? "basic" : "standard";
}

// Until AOL Mail supplies us with a proper API call, we will need to call this
// method to retrieve the email address; See bug 12445
AOLMailAccount.prototype._storeEmail =
function amAcct__storeEmail(aDocument) {
  // Email already stored, don't do it again
  if (this.getCustomParam(STR_EMAIL)) {
    return;
  }

  this._logger.debug("_storeEmail() called: no email stored");

  // Look for the email address in AOL's hidden tag
  var results = FlockSvcUtils.newResults();
  if (this._WebDetective
          .detect(CLASS_SHORT_NAME, "emailaddress", aDocument, results))
  {
    var email = results.getPropertyAsAString("email");
    this.setCustomParam(STR_EMAIL, email);
    this._logger.debug("stored email from tag: " + email);
    return;
  }

  // We couldn't find the hidden tag, so we fall back on our "api" call.
  // Scrape the contents of the Basic Version Compose page as it is currently
  // the only way to retrieve the user's email address correctly
  var url = this._WebDetective
                .getString(CLASS_SHORT_NAME, "basicComposeURL", "")
                .replace("%params%", "");

  var req = CC[XMLHTTPREQUEST_CONTRACTID].createInstance(CI.nsIXMLHttpRequest);
  req.open("GET", url, true);

  var inst = this;
  req.onreadystatechange = function aAcct__storeEmail_orsc(aEvent) {
    if (req.readyState == XMLHTTPREQUEST_READYSTATE_COMPLETED) {
      if (req.status / 100 == 2) {
        // We only want to store the email if we can find it, otherwise ignore
        var re = inst._WebDetective.getString(CLASS_SHORT_NAME, "email", "");
        var email = req.responseText.match(new RegExp(re, "im"));
        if (email && email[1]) {
          inst.setCustomParam(STR_EMAIL, email[1]);
          inst._logger.debug("stored email from API: " + email[1]);
        }
      }
    }
  };

  req.send(null);
}


/**************************************************************************
 * Component: Flock AOLMailAPI
 **************************************************************************/

function AOLMailAPI() {
  this.shortName = CLASS_SHORT_NAME;
  FlockSvcUtils.getLogger(this);
  this._logger.init("AOLMailAPI");
  FlockSvcUtils.getWD(this);
}

// void call(in AString aMethod,
//           in nsISupports aParams,
//           in flockIListener aListener);
AOLMailAPI.prototype.call =
function amAPI_call(aMethod, aParams, aListener) {
  this._logger.debug(".call('" + aMethod + "', ...)");
  if (!aMethod || !aParams || !aListener) {
    // The listener is required to report the results of the call
    this._logger.error("call: missing a parameter");
    return;
  }
  aListener.onStart(this, "call");

  // Get user id and account from param list
  var params = aParams.wrappedJSObject;
  var userId = params.userId
  var account = params.account;
  // Remove them from param list
  params.userId = null;
  params.account = null;

  var api = this;
  var authListener = {
    onStart: function call_authListener_onStart(aSubj, aTopic) {
      api._logger.debug("call_authListener_onStart()");
    },
    onSuccess: function call_authListener_onSuccess(aSubj, aTopic) {
      api._logger.debug("call_authListener_onSuccess()");

      var requestListener = {
        onStart: function call_requestListener_onStart(aSubj, aTopic) {
          api._logger.debug("call: requestListener.onStart()");
        },
        onSuccess: function call_requestListener_onSuccess(aResp, aTopic) {
          api._logger.debug("call: requestListener.onSuccess()");
          aListener.onSuccess(aResp, "call");
        },
        onError: function call_requestListener_onError(aResp, aTopic, aError) {
          api._logger.error("call: requestListener.onError('" + aError.errorCode
                                                              + "')");
          // If aResp is not null then this is an error in the response and
          // not the request.
          aListener.onError(aResp, "call", aError);
        }
      };

      var mailURL = api._WebDetective.getString(api.shortName, "mailURL", "")
                                     .replace("%method%", aMethod);
      var body = api._WebDetective.getString(api.shortName, "mailBody", "")
                                  .replace("%token%", api.token);
      // Append passed-in parameters to the body
      for (var prop in params) {
        if (params[prop]) {
          // The body will always have the devId and token in it, so it's all
          // right to prepend each parameter with an ampersand.
          body += "&" + prop + "=" + params[prop];
        }
      }
      // Send the request
      api._sendRequest("GET", mailURL, body, requestListener);
    },
    onError: function call_authListener_onError(aSubj, aTopic, aError) {
      api._logger.error("call_authListener_onError()");
      aListener.onError(api, "call", aError);
    }
  };
  // Ensure that API is authenticated
  this._authenticate(account, { user: userId }, authListener);
}

// flockIError getError(in AString aErrorCode);
AOLMailAPI.prototype.getError =
function amAPI_getError(aErrorCode) {
  this._logger.debug(".getError('" + aErrorCode +"')");
  var error = CC[FLOCK_ERROR_CONTRACTID].createInstance(CI.flockIError);
  error.serviceName = API_CLASS_NAME;
  if (aErrorCode) {
    error.errorCode = aErrorCode;
  } else {
    error.errorCode = CI.flockIError.WEBMAIL_UNKNOWN_ERROR;
  }
  return error;
}

// flockIError getHttpError(in AString aHttpErrorCode);
AOLMailAPI.prototype.getHttpError =
function amAPI_getHttpError(aHttpErrorCode) {
  this._logger.debug(".getHttpError('" + aHttpErrorCode +"')");
  var error = CC[FLOCK_ERROR_CONTRACTID].createInstance(CI.flockIError);
  error.serviceName = API_CLASS_NAME;
  if (aHttpErrorCode) {
    error.errorCode = aHttpErrorCode;
  } else {
    error.errorCode = CI.flockIError.WEBMAIL_UNKNOWN_ERROR;
  }
  return error;
}

// readonly attribute boolean isAuthenticated;
AOLMailAPI.prototype.isAuthenticated = false;

// void authenticate(in nsILoginInfo aCredentials, in flockIListener aListener);
AOLMailAPI.prototype.authenticate =
function amAPI_authenticate(aCredentials, aListener) {
  this._logger.debug(".authenticate(" + (aCredentials
                                         ? "'" + aCredentials.user + "', "
                                         : "")
                                      + "...)");
  if (aListener) {
    aListener.onStart(this, "authenticate");
  }

  if (this.isAuthenticated && this.token && this.expiresOn > Date.now()) {
    if (aListener) {
      aListener.onSuccess(this, "authenticate");
    }
    return;
  }

  // Need to get a token
  var api = this;
  var requestListener = {
    onStart: function auth_requestListener_onStart(aSubj, aTopic) {
      api._logger.debug("authenticate: requestListener.onStart()");
    },
    onSuccess: function auth_requestListener_onSuccess(aResponse, aTopic) {
      api._logger.debug("authenticate: requestListener.onSuccess()");
      // Get the token information from the response
      api.token = aResponse.data.token.a;
      api.expiresOn = (aResponse.data.token.expiresIn * 1000) + Date.now();
      api.isAuthenticated = true;
      if (aListener) {
        aListener.onSuccess(api, "authenticate");
      }
    },
    onError: function auth_requestListener_onError(aResponse, aTopic, aError) {
      api._logger.error("authenticate: requestListener.onError('"
                        + aError.errorCode + "')");
      // If aResponse is not null then this is an error in the response and
      // not the request.
      if (aResponse && aError.errorCode == MORE_AUTHENTICATION_REQUIRED_CODE) {
        // Additional authentication is required
        // :TODO: Handle this case if necessary
        if (aListener) {
          aListener.onError(api, "authenticate", aError);
        }
      } else {
        if (aListener) {
          aListener.onError(api, "authenticate", aError);
        }
      }
    }
  };

  var authURL = this._WebDetective.getString(this.shortName, "authURL", "")
                                  .replace("%method%", "getToken");
  var body = this._WebDetective.getString(this.shortName, "authBody", "")
                               .replace("%userId%", aCredentials
                                                    ? aCredentials.user
                                                    : "");
  this._sendRequest("POST", authURL, body, requestListener);
}

// void deauthenticate();
AOLMailAPI.prototype.deauthenticate =
function amAPI_deauthenticate() {
  this._logger.debug(".deauthenticate()");
  this.isAuthenticated = false;
  this.token = null;
}

// readonly attribute AString token;
AOLMailAPI.prototype.token = null;

// readonly attribute PRUint32 expiresOn;
AOLMailAPI.prototype.expiresOn = 0;

// void getNewMailCount(in AString aUid,
//                      in flockIWebmailAccount aAccount,
//                      in flockIListener aListener);
AOLMailAPI.prototype.getNewMailCount =
function amAPI_getNewMailCount(aUid, aAccount, aListener) {
  if (!aListener) {
    // The listener is required to report the results of the call
    this._logger.error("getNewMailCount: missing aListener parameter");
    return;
  }
  aListener.onStart(this, "getNewMailCount");
  var params = {
    wrappedJSObject: { "userId": aUid, "account": aAccount }
  };

  var api = this;
  var callListener = {
    onStart: function gnmc_callListener_onStart(aSubj, aTopic) {
      api._logger.debug("getNewMailCount: callListener.onStart()");
    },
    onSuccess: function gnmc_callListener_onSuccess(aResp, aTopic) {
      api._logger.debug("getNewMailCount: callListener.onSuccess()");
      var result = {
        wrappedJSObject: { "mailFolderCount": aResp.data.mailFolderCount }
      };
      aListener.onSuccess(result, "getNewMailCount");
    },
    onError: function gnmc_callListener_onError(aResp, aTopic, aError) {
      api._logger.error("getNewMailCount: callListener.onError('"
                        + aError.errorCode + "')");
      aListener.onError(aResp, "getNewMailCount", aError);
    }
  };

  this.call("newMailCount", params, callListener);
}

// void getNewMailList(in AString aUid,
//                     in AString aFolder,
//                     in AString aNumItems,
//                     in boolean aGetNewMailCount,
//                     in AString aLang,
//                     in AString aLocale,
//                     in flockIWebmailAccount aAccount,
//                     in flockIListener aListener);
AOLMailAPI.prototype.getNewMailList =
function amAPI_getNewMailList(aUid,
                              aFolder,
                              aNumItems,
                              aGetNewMailCount,
                              aLang,
                              aLocale,
                              aAccount,
                              aListener)
{
  if (!aListener) {
    // The listener is required to report the results of the call
    this._logger.error("getNewMailList: missing aListener parameter");
    return;
  }
  aListener.onStart(this, "getNewMailList");
  var params = {
    wrappedJSObject: { "userId": aUid,
                       "folder": aFolder,
                       "items": aNumItems,
                       "showCount": (aGetNewMailCount ? "yes" : null),
                       "lang": aLang,
                       "locale": aLocale,
                       "account": aAccount
                     }
  };

  var api = this;
  var callListener = {
    onStart: function gnml_callListener_onStart(aSubj, aTopic) {
      api._logger.debug("getNewMailList: callListener.onStart()");
    },
    onSuccess: function gnml_callListener_onSuccess(aResp, aTopic) {
      api._logger.debug("getNewMailList: callListener.onSuccess()");
      var result = {
        wrappedJSObject: { "mailList": aResp.data.mailList }
      };
      aListener.onSuccess(result, "getNewMailList");
    },
    onError: function gnml_callListener_onError(aResp, aTopic, aError) {
      api._logger.error("getNewMailList: callListener.onError('"
                        + aError.errorCode + "')");
      aListener.onError(aResp, "getNewMailList", aError);
    }
  };

  this.call("newMailList", params, callListener);
}

// A wrapper around authenticate() which will pull in the token and expiration
// info from the given account.
AOLMailAPI.prototype._authenticate =
function amAPI__authenticate(aAccount, aCredentials, aListener) {
  if (this.isAuthenticated && this.token && this.expiresOn > Date.now()) {
    if (aListener) {
      aListener.onSuccess(this, "_authenticate");
    }
    return;
  }

  // Grab the token information from the account if it is present
  if (aAccount) {
    this.token = aAccount.getCustomParam(STR_API_TOKEN);
    this.expiresOn = aAccount.getCustomParam(STR_API_EXPIRES_ON);
    // If we got valid info then mark the API as authenticated
    this.isAuthenticated = (this.token && this.expiresOn > Date.now());
  }
  // Get the user's login credentials if they were not supplied 
  var credentials;
  if (aCredentials) {
    credentials = aCredentials;
  } else if (aAccount) {
    var key = aAccount.urn.replace("account:", "service:")
                          .replace("flock:", "");
    credentials = aAccount._acUtils.getPassword(key);
  }
  var api = this;
  var authListener = {
    onStart: function a_authListener_onStart(aSubj, aTopic) {
      api._logger.debug("_authenticate: authListener.onStart()");
    },
    onSuccess: function a_authListener_onSuccess(aResp, aTopic) {
      api._logger.debug("_authenticate: authListener.onSuccess()");
      if (aAccount) {
        // Store the token info with the account
        aAccount.setCustomParam(STR_API_TOKEN, api.token);
        aAccount.setCustomParam(STR_API_EXPIRES_ON, api.expiresOn);
      }
      aListener.onSuccess(api, "_authenticate");
    },
    onError: function a_authListener_onError(aResp, aTopic, aError) {
      api._logger.error("_authenticate: authListener.onError('"
                        + aError.errorCode + "')");
      if (api.token) {
        // This token may be stale. Clear it and try again to get a fresh token.
        api.token = "";
        api.expiresOn = 0;
        api.authenticate(credentials, this);
      } else {
        aListener.onError(api, "_authenticate", aError);
      }
    }
  };
  this.authenticate(credentials, authListener);
}

// Send a request to the AOL server and wait for a reply
AOLMailAPI.prototype._sendRequest =
function amAPI__sendRequest(aMethod, aURL, aBody, aListener) {
  this._logger.debug("_sendRequest('" + aMethod + "', " +
                                  "'" + aURL + "', " +
                                  "'" + aBody + "')");

  aListener.onStart(this, "_sendRequest");
  if (aMethod != "POST" && aMethod != "GET") {
    aListener.onError(null, "_sendRequest", this.getHttpError("405"));
    return;
  }

  var url = aURL;
  if (aMethod == "GET") {
    url += "?" + aBody;
  }

  var req = CC[XMLHTTPREQUEST_CONTRACTID]
            .createInstance(CI.nsIXMLHttpRequest)
            .QueryInterface(CI.nsIJSXMLHttpRequest);
  req.open(aMethod, url, true);

  // "Referer" is misspelled on purpose here
  var referer = this._WebDetective.getString(this.shortName, "referer", "");
  req.setRequestHeader("Referer", referer);

  var body = null;
  if (aMethod == "POST") {
    req.setRequestHeader("Content-Type", "application/x-www-form-urlencoded");
    body = aBody;
  }

  var api = this;
  req.onreadystatechange = function aolapi_sr_orsc(aEvt) {
    if (req.readyState == XMLHTTPREQUEST_READYSTATE_COMPLETED) {
      try {
        api._logger.debug("._sendRequest() Status = " + req.status);
        api._logger.debug("._sendRequest() response = \n" + req.responseText);
        if (req.status / 100 == 2) {
          try {
            if (req.responseText) {
              var response = JSON.fromString(req.responseText).response;
              if (response.statusCode == 200) {
                aListener.onSuccess(response, "_sendRequest");
              } else {
                api._logger.debug("._sendRequest() response error - "
                                  + response.statusCode);
                // Include the response in the onError call so the listener
                // knows that this is an error in the response and not in the
                // initial request.
                aListener.onError(response,
                                  "_sendRequest",
                                  api.getHttpError(response.statusCode));
              }
            } else {
              api._logger.debug("._sendRequest - api returned no XML");
              aListener.onError(null,
                                "_sendRequest",
                                api.getError(CI.flockIError
                                               .WEBMAIL_API_INVALID_XML));
            }
          } catch (ex) {
            // Error parsing XML
            api._logger.error("._sendRequest() Parse error = " + ex);
            aListener.onError(null,
                              "_sendRequest",
                              api.getError(CI.flockIError
                                             .WEBMAIL_API_CANNOT_PARSE));
          }
        } else {
          // HTTP errors
          api._logger.debug("._sendRequest() HTTP Error - " + req.status);
          aListener.onError(null, "_sendRequest", api.getHttpError(req.status));
        }
      } catch (ex) {
        // XMLHTTPERROR (connection lost)
        api._logger.debug("._sendRequest() " + ex);
        aListener.onError(null, "_sendRequest", api.getHttpError("9001"));
      }
    }
  };

  req.send(body);
}


/**************************************************************************
 * XPCOM Support - Module Construction
 **************************************************************************/

// Create array of components.
var gComponentsArray = [AOLMailService];

// Generate a module for XPCOM to find.
var NSGetModule = FlockXPCOMUtils.generateNSGetModule(MODULE_NAME,
                                                      gComponentsArray);

/**************************************************************************
 * END XPCOM Support
 **************************************************************************/

