// BEGIN FLOCK GPL
//
// Copyright Flock Inc. 2005-2008
// http://flock.com
//
// This file may be used under the terms of the
// GNU General Public License Version 2 or later (the "GPL"),
// http://www.gnu.org/licenses/gpl.html
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// END FLOCK GPL

// Import this file with the following statement:
// CU.import("resource:///modules/webmailBackend.jsm");
var EXPORTED_SYMBOLS = ["webmailBackend"];

const CC = Components.classes;
const CI = Components.interfaces;
const CR = Components.results;
const CU = Components.utils;

CU.import("resource://gre/modules/JSON.jsm");

/**************************************************************************
 * Module: Webmail Back End
 **************************************************************************/

/* Note: This module depends on access to the following PRIVATE properties
         of webmail services and accounts being supported:
         - aAcct._WebDetective
*/

const STR_UNREAD = "unreadMessages";
const STR_WAITING_TO_SEND = "waitingToSend";
const AUTO_SET_PRIMARY_PREF = "flock.webmail.autoSetPrimaryAccount";

/**************************************************************************
 * Webmail Back End Public Interface
 **************************************************************************/

var webmailBackend = {};

webmailBackend.checkForPendingComposeMessage =
function WBE_checkForPendingComposeMessage(aAcct) {
  // Get pending compose URL
  var composeURL = aAcct.getCustomParam(STR_WAITING_TO_SEND);
  // Be sure the account is logged in
  if (composeURL && aAcct.isAuthenticated()) {
    // Clear pending URL from account
    aAcct.setCustomParam(STR_WAITING_TO_SEND, "");
    // Open pending compose message
    webmailBackend.openComposeMessageTab(aAcct, composeURL);
  }
};

webmailBackend.composeMessage =
function WBE_composeMessage(aAcct, aMailtoString) {
  // Get the compose URL from the account
  var composeURL = aAcct.getComposeURL(aMailtoString);
  if (!composeURL) {
    return false;
  }

  // If the account is logged in, open the compose page
  if (aAcct.isAuthenticated()) {
    return webmailBackend.openComposeMessageTab(aAcct, composeURL);
  }

  var acUtils = CC["@flock.com/account-utils;1"]
                .getService(CI.flockIAccountUtils);

  // Not logged in -- check if another account is logged in and we can log out
  // of it
  if (acUtils.ensureNoAuthenticatedAccountsBeforeLogin(aAcct.urn)) {
    // Save the compose URL to be sent once we've logged in
    aAcct.setCustomParam(STR_WAITING_TO_SEND, composeURL);

    // Open a login window
    var serviceId = acUtils.getServiceIDForAccountURN(aAcct.urn);
    var service = CC[serviceId].getService(CI.flockIMailWebService);
    var win = CC["@mozilla.org/appshell/window-mediator;1"]
              .getService(CI.nsIWindowMediator)
              .getMostRecentWindow("navigator:browser");
    win.openUILinkIn(service.loginURL, "tab");
  }

  return true;
};

webmailBackend.createEmptyMessageAsJSON =
function WBE_createEmptyMessageAsJSON() {
  var unreadMessages = {
    lastCheckedDate: Date.now(),
    newestMessageDate: 0,
    unreadMessageCount: 0,
    messages: []
  };
  return JSON.toString(unreadMessages);
};

webmailBackend.ensurePrimaryAccountExists =
function WBE_ensurePrimaryAccountExists() {
  var mostRecentPrimaryAccount = null;
  var mostRecentPrimaryDate = 0;
  // Get configured webmail accounts
  var acctsEnum = CC["@flock.com/account-utils;1"]
                  .getService(CI.flockIAccountUtils)
                  .getAccountsByInterface("flockIMailWebService");
  while (acctsEnum.hasMoreElements()) {
    var acct = acctsEnum.getNext()
                        .QueryInterface(CI.flockIMailWebServiceAccount);
    if (acct) {
      if (acct.isPrimary()) {
        // Found the primary account, so we're done
        return;
      }
      // Find most recent account to be set primary
      var lastPrimaryDate = acct.getCustomParam("lastPrimaryDate");
      // We include = here to catch the case where there are accounts but none
      // of them have ever been set to primary. This way, there will still be
      // a candidate to make primary.
      if (lastPrimaryDate >= mostRecentPrimaryDate) {
        mostRecentPrimaryDate = lastPrimaryDate;
        mostRecentPrimaryAccount = acct;
      }
    }
  }
  // There's no primary account so set most-recent as primary
  if (mostRecentPrimaryAccount) {
    mostRecentPrimaryAccount.setAsPrimary();
  }
};

webmailBackend.getPrimaryAccount =
function WBE_getPrimaryAccount() {
  // Get configured webmail accounts
  var acctsEnum = CC["@flock.com/account-utils;1"]
                  .getService(CI.flockIAccountUtils)
                  .getAccountsByInterface("flockIMailWebService");
  while (acctsEnum.hasMoreElements()) {
    var acct = acctsEnum.getNext()
                        .QueryInterface(CI.flockIMailWebServiceAccount);
    if (acct && acct.isPrimary()) {
      return acct;
    }
  }
  return null;
};

webmailBackend.handleRefreshedMessages =
function WBE_handleRefreshedMessages(aService, aURN, aMsgs) {
  // Get the mail account
  var account = aService.getAccount(aURN)
                        .QueryInterface(CI.flockIMailWebServiceAccount);
  if (account) {
    // Get old unread messages object
    var oldUnreads = JSON.fromString(account.getCustomParam(STR_UNREAD));
    // Get new unread messages object
    var newUnreads = JSON.fromString(aMsgs);

    // Check for new messages
    if (newUnreads.newestMessageDate > oldUnreads.newestMessageDate) {
      // Got new messages
      account.setCustomParam("hasNewMessages", true);
      // If this is the primary account, light the icon
      if (account.isPrimary()) {
        webmailBackend.lightIcon(true);
      }
    } else if (newUnreads.newestMessageDate < oldUnreads.newestMessageDate) {
      // new messages have been read
      account.setCustomParam("hasNewMessages", false);
      // If this is the primary account, unlight the icon
      if (account.isPrimary()) {
        webmailBackend.lightIcon(false);
      }
    }

    // Save new unreads
    account.setCustomParam(STR_UNREAD, aMsgs);
  }
};

webmailBackend.lightIcon =
function WBE_lightIcon(aHasNew) {
  CC["@mozilla.org/observer-service;1"]
    .getService(CI.nsIObserverService)
    .notifyObservers(null, "new-webmail-notification", aHasNew);
};

webmailBackend.onLogin =
function WBE_onLogin(aAcct) {
  var prefs = CC["@mozilla.org/preferences-service;1"]
              .getService(CI.nsIPrefBranch);
  // Do we need to set this as the primary account?
  if (prefs.getPrefType(AUTO_SET_PRIMARY_PREF) &&
      prefs.getBoolPref(AUTO_SET_PRIMARY_PREF))
  {
    aAcct.setAsPrimary();
  } else {
    webmailBackend.ensurePrimaryAccountExists();
  }

  var c_acct = aAcct._coop.get(aAcct.urn);
  // Do a refresh every time we log in
  c_acct.nextRefresh = new Date(0);
  // Ensure account is pollable
  c_acct.isPollable = true;
};

webmailBackend.openComposeMessageTab =
function WBE_openComposeMessageTab(aAcct, aComposeURL) {
  if (aAcct && aComposeURL) {
    // Get the current browser window
    var win = CC["@mozilla.org/appshell/window-mediator;1"]
              .getService(CI.nsIWindowMediator)
              .getMostRecentWindow("navigator:browser");
    if (win) {
      var browser = win.getBrowser();
      var onComposeLoad = function wbe_compose_onload() {
        // Do any handling necessary after the page has loaded
        browser.removeEventListener("load", onComposeLoad, true);
        aAcct.onComposeURLLoaded(aComposeURL, browser);
      };
      browser.addEventListener("load", onComposeLoad, true);
      // Open the url in a new tab on the window
      win.openUILinkIn(aComposeURL, "tab");
      return true;
    }
  }
  return false;
};

webmailBackend.parseMailParams =
function WBE_parseMailParams(aAcct, aComposeURL) {
  // Given a compose URL for an account, parse out the mailto fields.
  var paramList = {};
  if (aAcct && aComposeURL) {
    // Get query string
    var parts = aComposeURL.split("?");
    var query = parts[1] ? parts[1] : aComposeURL;
    // Remove anchor hooks
    query = query.split("#")[0];

    // Get name-value pairs
    var pairs = query.split("&");

    // If the mailto keys for this service differ from "to", "subject", "body",
    // "cc" and "bcc", we'll get their equivalents
    var cParams = aAcct._WebDetective
                       .getString(aAcct._shortName, "composeParams", "")
                       .split(",");
    // Fill a hash with these keys
    var paramSubs = {};
    for each (var keys in cParams) {
      var pair = keys.split("=");
      paramSubs[pair[1]] = pair[0];
    }

    // Go through the pairs and handle each as appropriate, adding them
    // to the corresponding parameter.
    for each (var pair in pairs) {
      parts = pair.split("=");
      var name = decodeURIComponent(parts.shift()).toLowerCase();
      // Look for substitute
      if (paramSubs[name]) {
        name = paramSubs[name]; 
      }
      var value = parts.join("=");
      if (name && value) {
        paramList[name] = decodeURIComponent(value);
      }
    }
  }

  return paramList;
};

// Returns true if the primary account is logged in and has new messages
webmailBackend.primaryAccountIsHot =
function WBE_primaryAccountIsHot() {
  var primaryAccount = webmailBackend.getPrimaryAccount();
  return (primaryAccount &&
          primaryAccount.isAuthenticated() &&
          primaryAccount.getCustomParam("hasNewMessages"));
};

webmailBackend.reformatMailParams =
function WBE_reformatMailParams(aAcct, aMailParams, aAddrSep) {
  // If aMailParams contains "mailto:", it will probably have a "?"
  // separating the recipient from the rest of the params.  Replace
  // with "to=" and "&", respectively.
  var params = "";
  if (aMailParams) {
    if (aMailParams.toLowerCase().indexOf("mailto:") == 0) {
      aMailParams = aMailParams.replace(/mailto:/i, "to=");
      aMailParams = aMailParams.replace(/\?/, "&");
    }

    // Get name-value pairs from parameters
    var pairs = aMailParams.split("&");

    // Go through the pairs and handle each as appropriate, adding them
    // to the corresponding parameter.
    var paramValues = {};
    for each (var pair in pairs) {
      var parts = pair.split("=");
      var name = decodeURIComponent(parts.shift()).toLowerCase();
      var value = parts.join("=");
      if (name && value) {
        switch (name) {
          case "to":
          case "cc":
          case "bcc":
            if (paramValues[name]) {
              // If there are mutiple address, separate by semi-colon
              paramValues[name] += aAddrSep + value;
            } else {
              paramValues[name] = value;
            }
            break;
          case "body":
            if (paramValues[name]) {
              // If there are multiple bodies, separate by newlines
              paramValues[name] += "\n" + value;
            } else {
              paramValues[name] = value;
            }
            break;
          case "subject":
            // If there are multiple subjects, use last one
            paramValues[name] = value;
            break;
        }
      }
    }

    // Replace the mailto keys with their webservice equivalents when
    // they differ from "to", "subject", "body", "cc" and "bcc"
    var cParams = aAcct._WebDetective
                       .getString(aAcct._shortName, "composeParams", "")
                       .split(",");
    // Fill a hash with these keys
    var paramSubs = {};
    for each (var keys in cParams) {
      var pair = keys.split("=");
      paramSubs[pair[0]] = pair[1];
    }

    function addParameterPair(aField) {
      if (paramValues[aField]) {
        params = params
               + (params ? "&" : "")
               + (paramSubs[aField] ? paramSubs[aField] : aField)
               + "="
               + paramValues[aField];
      }
    }

    // Build a new parameter string with these updated values
    addParameterPair("to");
    addParameterPair("cc");
    addParameterPair("bcc");
    addParameterPair("subject");
    addParameterPair("body");
  }

  return params;
};

webmailBackend.updateIcon =
function WBE_updateIcon() {
  webmailBackend.lightIcon(webmailBackend.primaryAccountIsHot());
};
